#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
ViT model definition.
Inspired from https://github.com/faustomorales/vit-keras/blob/master/vit_keras/vit.py.
"""
import keras

from quantizeml.layers import AddPositionEmbs, ClassToken, Add, ExtractToken
from quantizeml.models import load_model

from .model_vit import CONFIG_TI, CONFIG_S, CONFIG_B
from ..imagenet.imagenet_utils import IMAGENET_MEAN, IMAGENET_STD
from ..layer_blocks import norm_to_layer, transformer_block
from ..utils import fetch_file

BASE_WEIGHT_PATH = 'http://data.brainchip.com/models/AkidaV2/deit/'


def deit_imagenet(input_shape,
                  num_blocks,
                  hidden_size,
                  num_heads,
                  name,
                  mlp_dim,
                  patch_size=16,
                  classes=1000,
                  dropout=0.1,
                  include_top=True,
                  distilled=False,
                  norm='LN',
                  last_norm='LN',
                  softmax='softmax',
                  act="GeLU"):
    """Build a DeiT model.

    Args:
        input_shape (tuple): image shape tuple
        num_blocks (int): the number of transformer blocks to use.
        hidden_size (int): the number of filters to use
        num_heads (int): the number of transformer heads
        name (str): the model name
        mlp_dim (int): the number of dimensions for the MLP output in the transformers.
        patch_size (int, optional): the size of each patch (must fit evenly in image size). Defaults
            to 16.
        classes (int, optional): number of classes to classify images into, only to be specified if
            `include_top` is True. Defaults to 1000.
        dropout (float, optional): fraction of the units to drop for dense layers. Defaults to 0.1.
        include_top (bool, optional): whether to include the final classifier head. If False,
            the output will correspond to that of the transformer. Defaults to True.
        distilled (bool, optional): Build model append a distilled token. Defaults to False.
        norm (str, optional): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to
            choose from LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization
            or LayerMadNormalization layers respectively in the model. Defaults to 'LN'.
        last_norm (str, optional): string that values in ['LN', 'BN']
            and that allows to choose from LayerNormalization or
            BatchNormalization in the classifier network. Defaults to 'LN'.
        softmax (str, optional): string with values in ['softmax', 'softmax2']
            that allows to choose between softmax and softmax2 in MHA. Defaults
            to 'softmax'.
        act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation in MLP block. Defaults to 'GeLU'.
    """
    assert ((input_shape[0] % patch_size == 0) and
            (input_shape[1] % patch_size == 0)), "image size must be a multiple of patch_size"

    if last_norm not in ('LN', 'BN'):
        raise NotImplementedError("last_norm should be in ['LN', 'BN']"
                                  f"but received {norm}.")

    # Normalize image adding rescaling layer
    x = keras.layers.Input(shape=input_shape, name="input")
    scale = list(1.0 / 255 / std for std in IMAGENET_STD)
    offset = list(-mean / std for mean, std in zip(IMAGENET_MEAN, IMAGENET_STD))
    y = keras.layers.Rescaling(scale=scale, offset=offset, name="Rescale")(x)

    # Build model
    y = keras.layers.Conv2D(
        filters=hidden_size,
        kernel_size=patch_size,
        strides=patch_size,
        padding="valid",
        name="Embedding",
        kernel_initializer=keras.initializers.TruncatedNormal(stddev=0.02),
        bias_initializer="zeros",
    )(y)
    y = keras.layers.Reshape((y.shape[1] * y.shape[2], hidden_size))(y)
    if distilled:
        y = ClassToken(name="DistToken")(y)
    y = ClassToken(name="ClassToken")(y)
    y = AddPositionEmbs(name="Transformer/PosEmbed")(y)
    for n in range(num_blocks):
        y, _ = transformer_block(
            y,
            num_heads=num_heads,
            hidden_size=hidden_size,
            mlp_dim=mlp_dim,
            dropout=dropout,
            name=f"Transformer/EncoderBlock_{n}",
            norm=norm,
            softmax=softmax,
            mlp_act=act,
        )

    # Include classification head
    if include_top:
        yt = norm_to_layer(last_norm)(
            epsilon=1e-6, name="Transformer/EncoderNorm")(y)
        y = ExtractToken(token=0, name="ExtractToken")(yt)
        y = keras.layers.Dense(classes, name="Head")(y)
        if distilled:
            yd = ExtractToken(token=1, name="ExtractToken_Dist")(yt)
            yd = keras.layers.Dense(classes, name="DistHead")(yd)
            y = Add(name="Add", average=True)([y, yd])

    # Add distilled flag
    model = keras.models.Model(inputs=x, outputs=y, name=name)
    model.isdistilled = distilled
    return model


def deit_ti16(input_shape=(224, 224, 3),
              classes=1000,
              distilled=False,
              norm='LN',
              last_norm='LN',
              softmax='softmax',
              act='GeLU',
              include_top=True):
    """Build DeiT-Tiny.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        distilled (bool, optional): build model appending a distilled token. Defaults to False.
        norm (str, optional): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to
            choose from LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization
            or LayerMadNormalization layers respectively in the model. Defaults to 'LN'.
        last_norm (str, optional): string that values in ['LN', 'BN']
            and that allows to choose from LayerNormalization or
            BatchNormalization in the classifier network. Defaults to 'LN'.
        softmax (str, optional): string with values in ['softmax', 'softmax2'] that allows to choose
            between softmax and softmax2 in attention block. Defaults to 'softmax'.
        act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation inside MLP. Defaults to 'GeLU'.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return deit_imagenet(
        name="deit-tiny",
        input_shape=input_shape,
        classes=classes,
        distilled=distilled,
        norm=norm,
        last_norm=last_norm,
        act=act,
        softmax=softmax,
        include_top=include_top,
        **CONFIG_TI,
    )


def bc_deit_ti16(input_shape=(224, 224, 3), classes=1000, distilled=False, include_top=True,
                 num_blocks=12):
    """Build DeiT-Tiny, changing all LN by LMN, using softmax2 and ReLU8.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        distilled (bool, optional): build model appending a distilled token. Defaults to False.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.
        num_blocks (int, optional): the number of transformer blocks to use. Defaults to 12.

    Returns:
        keras.Model: the requested model
    """
    config_ti = CONFIG_TI.copy()
    config_ti["num_blocks"] = num_blocks

    return deit_imagenet(
        name="deit-tiny",
        input_shape=input_shape,
        classes=classes,
        distilled=distilled,
        norm="LMN",
        last_norm="BN",
        softmax="softmax2",
        act="ReLU8",
        include_top=include_top,
        **config_ti,
    )


def bc_deit_dist_ti16_imagenet_pretrained():
    """ Helper method to retrieve a `bc_deit_dist_ti16` model that was trained on ImageNet dataset.

    Returns:
        keras.Model, dict: a Keras Model instance and the quantization configuration as a JSON
        formatted dict.
    """
    model_name = 'bc_deit_dist_ti16_224_i8_w4_a4.h5'
    file_hash = '0975fa986c8d026929568ae47022b153a39d9bb6327cb2aa8b0a38b3c03f6025'
    model_path = fetch_file(BASE_WEIGHT_PATH + model_name,
                            fname=model_name,
                            file_hash=file_hash,
                            cache_subdir='models')
    return load_model(model_path)


def deit_s16(input_shape=(224, 224, 3), classes=1000, distilled=False, include_top=True):
    """Build DeiT-Small.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        distilled (bool, optional): build model appending a distilled token. Defaults to False.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return deit_imagenet(
        name="deit-small",
        input_shape=input_shape,
        classes=classes,
        distilled=distilled,
        include_top=include_top,
        **CONFIG_S,
    )


def deit_b16(input_shape=(224, 224, 3), classes=1000, distilled=False, include_top=True):
    """Build DeiT-B16.

    Args:
        input_shape (tuple, optional): input shape. Defaults to (224, 224, 3).
        classes (int, optional): number of classes. Defaults to 1000.
        distilled (bool, optional): build model appending a distilled token. Defaults to False.
        include_top (bool, optional): whether to include the final classifier network.
            Defaults to True.

    Returns:
        keras.Model: the requested model
    """
    return deit_imagenet(
        name="deit-base",
        input_shape=input_shape,
        classes=classes,
        distilled=distilled,
        include_top=include_top,
        **CONFIG_B,
    )



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for DVS models.
"""

import os
import argparse
import numpy as np
import tensorflow as tf

from keras.utils.np_utils import to_categorical

from cnn2snn import load_quantized_model, convert

from ..cyclic_lr import CyclicLR
from ..training import (get_training_parser, compile_model, evaluate_model, evaluate_akida_model,
                        print_history_stats, RestoreBest, save_model)
from ..utils import fetch_file


def get_data(dataset, batch_size):
    """ Loads data.

    Dataset parameters for DVSDataGenerator are set from DVS pickles.

    Args:
        dataset (str): name of the dataset to load, either 'dvs_gesture' or
            'samsung_handy'.
        batch_size (int): the batch size

    Returns:
        tf.dataset, tf.dataset, int: train dataset, test dataset and number of
        samples.
    """
    # Load pre-processed dataset
    train_file = fetch_file(os.path.join('http://data.brainchip.com/dataset-mirror',
                                         dataset, dataset + '_preprocessed_train.npz'),
                            fname=dataset + '_preprocessed_train.npz',
                            cache_subdir=os.path.join('datasets', dataset))

    test_file = fetch_file(os.path.join('http://data.brainchip.com/dataset-mirror',
                                        dataset, dataset + '_preprocessed_test.npz'),
                           fname=dataset + '_preprocessed_test.npz',
                           cache_subdir=os.path.join('datasets', dataset))
    train_data = np.load(train_file)
    x_train = train_data['x_train']
    y_train = to_categorical(train_data['y_train'])

    test_data = np.load(test_file)
    x_test = test_data['x_test']
    y_test = to_categorical(test_data['y_test'])

    train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train))
    test_dataset = tf.data.Dataset.from_tensor_slices((x_test, y_test))

    data_augmentation = tf.keras.Sequential([
        tf.keras.layers.RandomRotation(0.05, interpolation='nearest'),
    ])

    train_dataset = train_dataset.shuffle(
        buffer_size=1024).batch(batch_size).map(
            lambda x, y: (data_augmentation(x, training=True), y),
            num_parallel_calls=tf.data.AUTOTUNE).prefetch(
                buffer_size=tf.data.AUTOTUNE)
    test_dataset = test_dataset.batch(batch_size).prefetch(
        buffer_size=tf.data.AUTOTUNE)

    return train_dataset, test_dataset, x_train.shape[0]


def train_model(model, train_ds, test_ds, batch_size, max_learning_rate, epochs,
                num_samples):
    """ Trains the model.

    Args:
        model (keras.Model): the model to train
        train_ds (tf.dataset): train dataset
        test_ds (tf.dataset): test dataset
        batch_size (int): the batch size
        max_learning_rate (float): learning rate maximum value
        epochs (int): the number of epochs
        num_samples (int): number of samples
    """
    n_iterations = np.round(num_samples / batch_size)

    # Define training callbacks
    callbacks = []

    lr_scheduler = CyclicLR(base_lr=max_learning_rate * 0.01,
                            max_lr=max_learning_rate,
                            step_size=4 * n_iterations,
                            mode='triangular2')
    callbacks.append(lr_scheduler)

    # Model checkpoints (save best model and retrieve it when training is complete)
    restore_model = RestoreBest(model)
    callbacks.append(restore_model)

    history = model.fit(train_ds,
                        epochs=epochs,
                        callbacks=callbacks,
                        validation_data=test_ds)
    print_history_stats(history)


def main():
    """ Entry point for script and CLI usage.
    """
    global_parser = argparse.ArgumentParser(add_help=False)
    global_parser.add_argument("-d",
                               "--dataset",
                               type=str,
                               default='dvs_gesture',
                               choices=['dvs_gesture', 'samsung_handy'],
                               help="Dataset name (defaut=dvs_gesture)")

    parsers = get_training_parser(batch_size=32,
                                  tune=True,
                                  global_parser=global_parser)
    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    learning_rate = 0.001
    if args.action == 'tune':
        learning_rate = 0.0001

    compile_model(model, learning_rate=learning_rate)

    # Load data
    train_ds, test_ds, num_samples = get_data(args.dataset, args.batch_size)

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    # Train model
    if args.action in ["train", "tune"]:
        train_model(model, train_ds, test_ds, args.batch_size, learning_rate,
                    args.epochs, num_samples)
        save_model(model, args.model, args.savemodel, args.action)
    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model, input_is_image=False)
            preds, labels = evaluate_akida_model(model_ak, test_ds, 'softmax')
            accuracy = (np.squeeze(np.argmax(preds, 1)) == np.argmax(labels,
                                                                     1)).mean()
            print(f"Akida accuracy: {accuracy}")
        else:
            evaluate_model(model, test_ds)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Training script for MNIST models.
"""

import os
from keras.datasets import mnist
from keras.losses import SparseCategoricalCrossentropy

from cnn2snn import load_quantized_model, convert

import numpy as np

from ..training import (get_training_parser, compile_model, evaluate_model, print_history_stats,
                        save_model, RestoreBest)


def get_data():
    """ Loads MNIST data.

    Returns:
        tuple: train data, train labels, test data and test labels
    """
    # The data, shuffled and split between train and test sets
    (x_train, y_train), (x_test, y_test) = mnist.load_data()

    # Normalize data
    x_train = x_train.astype("float32")
    x_train = np.reshape(x_train, (-1, 28, 28, 1))

    x_test = x_test.astype("float32")
    x_test = np.reshape(x_test, (-1, 28, 28, 1))

    return x_train, y_train, x_test, y_test


def train_model(model, x_train, y_train, x_test, y_test, batch_size, epochs):
    """ Trains the model without the distillation.

    Args:
        model (keras.Model): the model to train
        x_train (numpy.ndarray): train data
        y_train (numpy.ndarray): train labels
        x_test (numpy.ndarray): test data
        y_test (numpy.ndarray): test labels
        batch_size (int): the batch size
        epochs (int): the number of epochs
    """
    history = model.fit(x_train, y_train, batch_size, epochs, validation_data=(x_test, y_test),
                        callbacks=[RestoreBest(model)])
    print_history_stats(history)


def main():
    """ Entry point for script and CLI usage.
    """
    parsers = get_training_parser(batch_size=128, tune=True, global_batch_size=False)
    args = parsers[0].parse_args()

    # Load the source model
    model = load_quantized_model(args.model)

    # Compile model
    lr = 1e-4
    if args.action == 'tune':
        lr = 3e-5
    compile_model(model, learning_rate=lr, loss=SparseCategoricalCrossentropy(from_logits=True))

    # Load data
    x_train, y_train, x_test, y_test = get_data()

    # Disable QuantizeML assertions
    os.environ["ASSERT_ENABLED"] = "0"

    # Train model
    if args.action in ["train", "tune"]:
        train_model(model, x_train, y_train, x_test, y_test, args.batch_size, args.epochs)
        save_model(model, args.model, args.savemodel, args.action)
    elif args.action == "eval":
        # Evaluate model accuracy
        if args.akida:
            model_ak = convert(model)
            accuracy = model_ak.evaluate(x_test.astype('uint8'), y_test)
            print(f"Accuracy: {accuracy}")
        else:
            evaluate_model(model, x_test, y=y_test)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2022 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Utilities for akida_models package.
"""

import os
import urllib
import time
import numpy as np
import warnings
from pathlib import Path

from six.moves.urllib.parse import urlsplit

import tensorflow as tf
import tensorflow_addons as tfa
from keras.utils import io_utils
from keras.utils.data_utils import validate_file, _extract_archive
from keras.utils.generic_utils import Progbar
from keras.callbacks import TensorBoard

from cnn2snn import load_quantized_model
from quantizeml.models import load_model as qml_load_model


def fetch_file(origin, fname=None, file_hash=None, cache_subdir="datasets", extract=False,
               cache_dir=None):
    """ Downloads a file from a URL if it is not already in the cache.

    Reimplements `keras.utils.get_file` without raising an error when detecting a file_hash
    mismatch (it will just re-download the model).

    Args:
        origin (str): original URL of the file.
        fname (str, optional): name of the file. If an absolute path `/path/to/file.txt` is
            specified the file will be saved at that location. If `None`, the name of the file at
            `origin` will be used. Defaults to None.
        file_hash (str, optional): the expected hash string of the file after download. Defaults to
            None.
        cache_subdir (str, optional): subdirectory under the Keras cache dir where the file is
            saved. If an absolute path `/path/to/folder` is specified the file will be saved at that
            location. Defaults to 'datasets'.
        extract (bool, optional): True tries extracting the file as an Archive, like tar or zip.
            Defaults to False.
        cache_dir (str, optional): location to store cached files, when None it defaults to the
            default directory `~/.keras/`. Defaults to None.

    Returns:
        str: path to the downloaded file
    """
    if cache_dir is None:
        cache_dir = os.path.join(os.path.expanduser("~"), ".keras")

    datadir_base = os.path.expanduser(cache_dir)
    if not os.access(datadir_base, os.W_OK):
        datadir_base = os.path.join("/tmp", ".keras")
    datadir = os.path.join(datadir_base, cache_subdir)
    os.makedirs(datadir, exist_ok=True)

    fname = io_utils.path_to_string(fname)
    if not fname:
        fname = os.path.basename(urlsplit(origin).path)
        if not fname:
            raise ValueError(f"Can't parse the file name from the origin provided: '{origin}'."
                             "Please specify the `fname` as the input param.")

    fpath = os.path.join(datadir, fname)

    download = False
    if os.path.exists(fpath):
        # File found, verify integrity if a hash was provided.
        if file_hash is not None and not validate_file(fpath, file_hash):
            io_utils.print_msg("A local file was found, but it seems to be incomplete or outdated"
                               "because the file hash does not match the original value of "
                               f"{file_hash} so we will re-download the data.")
            download = True
    else:
        download = True

    if download:
        io_utils.print_msg(f"Downloading data from {origin}.")

        class DLProgbar:
            """Manage progress bar state for use in urlretrieve."""

            def __init__(self):
                self.progbar = None
                self.finished = False

            def __call__(self, block_num, block_size, total_size):
                if not self.progbar:
                    if total_size == -1:
                        total_size = None
                    self.progbar = Progbar(total_size)
                current = block_num * block_size
                if current < total_size:
                    self.progbar.update(current)
                elif not self.finished:
                    self.progbar.update(self.progbar.target)
                    self.finished = True

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                urllib.request.urlretrieve(origin, fpath, DLProgbar())
            except urllib.error.HTTPError as e:
                raise Exception(error_msg.format(origin, e.code, e.msg))
            except urllib.error.URLError as e:
                raise Exception(error_msg.format(origin, e.errno, e.reason))
        except (Exception, KeyboardInterrupt):
            if os.path.exists(fpath):
                os.remove(fpath)
            raise

    if extract:
        _extract_archive(fpath, datadir)

    return fpath


def load_model(model_path):
    """Combine the cnn2snn.load_quantized_model and quantizeml.load_model

    Args:
        model_path (str): model path

    Returns:
        keras.Model: the load model
    """
    try:
        model = load_quantized_model(model_path)
    except Exception:
        try:
            model = qml_load_model(model_path, custom_layers={'GELU': tfa.layers.GELU})
        except Exception as e:
            raise e.__class__('Cannot load provided model.')
    return model


def get_tensorboard_callback(out_dir, histogram_freq=1, prefix=''):
    """Build a Tensorboard call, pointing to the output directory

    Args:
        out_dir (str): parent directory of the folder to create
        histogram_freq (int, optional): frequency to export logs. Defaults to 1.
        prefix (str, optional): prefix name. Defaults to ''.
    """
    def _create_log_dir(out_dir, prefix=''):
        if len(prefix) != 0 and not prefix.endswith('_'):
            prefix += '_'
        base_name = prefix + time.strftime('%Y_%m_%d.%H_%M_%S', time.localtime())
        log_dir = os.path.join(out_dir, base_name)

        print('Saving tensorboard and checkpoint information to:', log_dir)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
            print('Directory', log_dir, 'created ...')
        else:
            print('Directory', log_dir, 'already exists ...')
        return log_dir

    log_dir = _create_log_dir(out_dir, prefix)
    file_writer = tf.summary.create_file_writer(log_dir + "/metrics")
    file_writer.set_as_default()
    return TensorBoard(log_dir=log_dir,
                       histogram_freq=histogram_freq,
                       update_freq='epoch',
                       write_graph=False,
                       profile_batch=0)


def apply_weights_to_model(model, weights, verbose=True):
    """Loads weights from a dictionary and apply it to a model.

    Go through the dictionary of weights, find the corresponding variable
    in the model and partially load its weights.

    Args:
        model (keras.Model): the model to update
        weights (dict): the dictionary of weights
        verbose (bool, optional): if True, throw warning messages if a dict
        item is not found in the model. Defaults to True.
    """
    if len(weights) == 0:
        warnings.warn("There is no weight to apply to the model.")
        return

    # Go through the dictionary of weights with each item
    for key, value in weights.items():
        value_applied = False
        for dest_var in model.variables:
            if key == dest_var.name:
                # Apply the current item value
                dest_var.assign(value)
                value_applied = True
                break
        if not value_applied and verbose:
            warnings.warn(f"Variable '{key}' not found in the model.")


def load_weights(model, weights_path):
    """Loads weights from a npz file and apply it to a model.

    Go through the dictionary of weights of the npz file, find the
    corresponding variable in the model and partially load its weights.

    Args:
        model (keras.Model): the model to update
        weights_path (str): the path of the npz file to load
    """
    # Check the npz file validity
    path = Path(weights_path)
    if not path.is_file():
        raise ValueError(f"File `{weights_path}` not found.")

    # Open the npz file
    weights_dict = np.load(weights_path)

    # Apply the weights to the model
    apply_weights_to_model(model, weights_dict)


def save_weights(model, weights_path):
    """Save model weights on an npz file.

    Takes a model and save the weights of all its layers into an npz file.

    Args:
        model (keras.Model): the model to save its weights
        weights_path (str): the path of the npz file to save
    """
    weights_dict = {}
    for var in model.variables:
        weights_dict[var.name] = var

    np.savez(weights_path, **weights_dict)



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Layers blocks definitions.
"""
from functools import partial

from keras.layers import (BatchNormalization, ReLU, Conv2D, DepthwiseConv2D, SeparableConv2D, Dense,
                          MaxPool2D, GlobalAvgPool2D, LayerNormalization, Dropout, Add,
                          Conv2DTranspose, Reshape)
from keras.activations import swish
from keras.initializers import TruncatedNormal
from tensorflow_addons.layers import GELU, GroupNormalization

from quantizeml.layers import (Attention, LayerMadNormalization, DepthwiseConv2DTranspose,
                               ExtractToken)


def _add_pooling_layer(x, pooling_type, pool_size, padding, layer_base_name):
    """Add a pooling layer in the graph.

    From an input tensor 'x', the function returns the output tensor after
    a pooling layer defined by 'pooling_type'.

    Args:
        x (tf.Tensor): the input tensor
        pooling_type (str): type of pooling among the following: 'max' or 'global_avg'.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        padding (str): one of "valid" or "same" (case-insensitive).
        layer_base_name (str): base name for the pooling layer.

    Returns:
        tf.Tensor: an output tensor after pooling
    """
    if pooling_type == 'max':
        return MaxPool2D(pool_size=pool_size,
                         padding=padding,
                         name=layer_base_name + '/maxpool')(x)
    if pooling_type == 'global_avg':
        return GlobalAvgPool2D(name=layer_base_name + '/global_avg')(x)
    raise ValueError("'pooling_type' argument must be 'max' or 'global_avg'.")


def conv_block(inputs,
               filters,
               kernel_size,
               pooling=None,
               post_relu_gap=False,
               pool_size=(2, 2),
               add_batchnorm=False,
               add_activation=True,
               **kwargs):
    """Adds a convolutional layer with optional layers in the following order:
    max pooling, batch normalization, activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(rows, cols, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution kernel.
            Can be a single integer to specify the same value for
            all spatial dimensions.
        pooling (str): add a pooling layer of type 'pooling' among the values 'max' or 'global_avg',
            with pooling size set to pool_size. If 'None', no pooling will be added.
        post_relu_gap (bool): when pooling is 'global_avg', indicates if the pooling comes before or
            after ReLU activation. Defaults to False.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        **kwargs: arguments passed to the keras.Conv2D layer, such as
            strides, padding, use_bias, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of conv2D block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in conv_block must be None.")
    if 'dilation_rate' in kwargs and kwargs['dilation_rate'] not in [1, [1, 1], (1, 1)]:
        raise ValueError("Keyword argument 'dilation_rate' is not supported in conv_block.")

    conv_layer = Conv2D(filters, kernel_size, **kwargs)
    x = conv_layer(inputs)

    if pooling == 'max' or (pooling == 'global_avg' and not post_relu_gap):
        x = _add_pooling_layer(x, pooling, pool_size, conv_layer.padding, conv_layer.name)

    if add_batchnorm:
        x = BatchNormalization(name=conv_layer.name + '/BN')(x)

    if add_activation:
        x = ReLU(6.0, name=conv_layer.name + '/relu')(x)

    if post_relu_gap and pooling == 'global_avg':
        x = _add_pooling_layer(x, pooling, pool_size, conv_layer.padding, conv_layer.name)
    return x


def separable_conv_block(inputs,
                         filters,
                         kernel_size,
                         strides=1,
                         padding="same",
                         use_bias=True,
                         pooling=None,
                         post_relu_gap=False,
                         pool_size=(2, 2),
                         add_batchnorm=False,
                         add_activation=True,
                         fused=True,
                         name=None,
                         kernel_initializer='glorot_uniform',
                         pointwise_regularizer=None):
    """Adds a separable convolutional layer with optional layers in the
    following order: global average pooling, max pooling, batch normalization,
    activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(height, width, channels)`
        filters (int): the dimensionality of the output space
            (i.e. the number of output filters in the pointwise convolution).
        kernel_size (int or tuple of 2 integers): specifying the
            height and width of the 2D convolution window. Can be a single
            integer to specify the same value for all spatial dimensions.
        strides (int or tuple of 2 integers, optional): strides of the depthwise convolution.
            Defaults to 1.
        padding (str, optional): padding mode for the depthwise convolution. Defaults to 'same'.
        use_bias (bool, optional): whether the layer uses a bias vector. Defaults to True.
        pooling (str): add a pooling layer of type 'pooling' among the values 'max', or
            'global_avg', with pooling size set to pool_size. If 'None', no pooling will be added.
        post_relu_gap (bool): when pooling is 'global_avg', indicates if the pooling comes before or
            after ReLU activation. Defaults to False.
        pool_size (int or tuple of 2 integers): factors by which to
            downscale (vertical, horizontal). (2, 2) will halve the input in
            both spatial dimension. If only one integer is specified, the same
            window length will be used for both dimensions.
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        fused (bool): If True use a SeparableConv2D layer otherwise use a
            DepthwiseConv2D + Conv2D layers. Defaults to True.
        name (str, optional): name of the layer. Defaults to None.
        kernel_initializer (keras.initializer, optional): initializer for both kernels. Defaults to
            'glorot_uniform'.
        pointwise_regularizer (keras.regularizers, optional): regularizer function applied to the
            pointwise kernel matrix. Defaults to None.

    Returns:
        tf.Tensor: output tensor of separable conv block.
    """
    if name:
        dw_name = "dw_" + name
        pw_name = "pw_" + name

    else:
        dw_name = pw_name = None
    # if fused set a SeparableConv2D layer
    if fused:
        sep_conv_layer = SeparableConv2D(filters, kernel_size, strides=strides, padding=padding,
                                         use_bias=use_bias,
                                         depthwise_initializer=kernel_initializer,
                                         pointwise_initializer=kernel_initializer,
                                         pointwise_regularizer=pointwise_regularizer,
                                         name=name)
        x = sep_conv_layer(inputs)
        main_layer_name = sep_conv_layer.name
    # if not fused set a DepthwiseConv2D + Conv2D layer (the Conv2D applies a Pointwise convolution)
    else:
        depth_conv_layer = DepthwiseConv2D(kernel_size, strides=strides, padding=padding,
                                           use_bias=False,
                                           depthwise_initializer=kernel_initializer,
                                           name=dw_name)
        point_conv_layer = Conv2D(filters, (1, 1), use_bias=use_bias,
                                  kernel_initializer=kernel_initializer,
                                  kernel_regularizer=pointwise_regularizer,
                                  name=pw_name)

        x = depth_conv_layer(inputs)
        x = point_conv_layer(x)

        main_layer_name = point_conv_layer.name

    if pooling == 'max' or (pooling == 'global_avg' and not post_relu_gap):
        x = _add_pooling_layer(x, pooling, pool_size, padding, main_layer_name)

    if add_batchnorm:
        x = BatchNormalization(name=main_layer_name + '/BN')(x)

    if add_activation:
        x = ReLU(6.0, name=main_layer_name + '/relu')(x)

    if post_relu_gap and pooling == 'global_avg':
        x = _add_pooling_layer(x, pooling, pool_size, padding, main_layer_name)
    return x


def dense_block(inputs,
                units,
                add_batchnorm=False,
                add_activation=True,
                **kwargs):
    """Adds a dense layer with optional layers in the following order:
    batch normalization, activation.

    Args:
        inputs (tf.Tensor): Input tensor of shape `(rows, cols, channels)`
        units (int): dimensionality of the output space
        add_batchnorm (bool): add a BatchNormalization layer
        add_activation (bool): add a ReLU layer
        **kwargs: arguments passed to the Dense layer, such as
            use_bias, kernel_initializer, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of the dense block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in dense_block must be None.")

    dense_layer = Dense(units, **kwargs)
    x = dense_layer(inputs)

    if add_batchnorm:
        x = BatchNormalization(name=dense_layer.name + '/BN')(x)

    if add_activation:
        x = ReLU(6.0, name=dense_layer.name + '/relu')(x)

    return x


def act_to_layer(act, **kwargs):
    """ Get activation layer from string.

    This is needed because one cannot serialize a class in layer.get_config, the string is thus
    serialized instead.

    Args:
        act (str): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to choose from
            GeLU, ReLUx or swish activation inside MLP.

    Returns:
        keras.layers: the activation layer class
    """
    if act == 'GeLU':
        act_funct = GELU(**kwargs)
    elif 'ReLU' in act:
        if act == 'ReLU':
            max_value = None
        else:
            try:
                max_value = float(act[4:])
            except ValueError:
                raise ValueError("ReLU must be in the form 'ReLUx', where x is the max-value")
        act_funct = ReLU(max_value=max_value, **kwargs)
    elif act == 'swish':
        act_funct = swish
    else:
        raise NotImplementedError(
            f"act should be in ['GeLU', 'ReLUx', 'swish'] but received {act}.")

    return act_funct


def norm_to_layer(norm):
    """ Get normalization layer from string.

    This is needed because one cannot serialize a class in layer.get_config, the string is thus
    serialized instead.

    Args:
        norm (str): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to choose from
            LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization or
            LayerMadNormalization layers respectively in the model.

    Returns:
        keras.layers: the normalization layer class
    """
    if norm == 'LN':
        norm_funct = LayerNormalization
    elif norm == 'GN1':
        norm_funct = partial(GroupNormalization, groups=1)
    elif norm == 'BN':
        norm_funct = BatchNormalization
    elif norm == 'LMN':
        norm_funct = LayerMadNormalization
    else:
        raise NotImplementedError("norm should be in ['LN', 'GN1', 'BN', 'LMN']"
                                  f" but received {norm}.")

    return norm_funct


def mlp_block(inputs, mlp_dim, dropout, name, mlp_act="GeLU"):
    """ MLP block definition.

    Args:
        inputs (tf.Tensor): inputs
        mlp_dim (int): number of units in the first dense layer
        dropout (float): dropout rate
        name (str): used as a base name for the layers in the block
        mlp_act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation. Defaults to "GeLU".

    Returns:
        tf.Tensor: MLP block outputs
    """
    initializer = {
        "kernel_initializer": TruncatedNormal(stddev=0.02),
        "bias_initializer": "zeros",
    }
    x = Dense(
        mlp_dim,
        name=f"{name}/Dense_0",
        **initializer,
    )(inputs)
    x = act_to_layer(mlp_act, name=f"{name}/activation")(x)
    x = Dropout(dropout)(x)
    x = Dense(
        inputs.shape[-1],
        name=f"{name}/Dense_1",
        **initializer,
    )(x)
    outputs = Dropout(dropout)(x)
    return outputs


def multi_head_attention(x, num_heads, hidden_size, name, softmax="softmax"):
    """Multi-head attention block definition.

    Args:
        x (tf.Tensor): inputs
        num_heads (int): the number of attention heads
        hidden_size (int): query, key and value dense layers representation size (units)
        name (str): used as a base name for the layers in the block
        softmax (str, optional): string with values in ['softmax', 'softmax2'] that allows to choose
            between softmax and softmax2 activation. Defaults to 'softmax'.

    Raises:
        ValueError: if hidden_size is not a multiple of num_heads

    Returns:
        (tf.Tensor, tf.Tensor): block outputs and attention softmaxed scores
    """
    if hidden_size % num_heads != 0:
        raise ValueError(
            f"Embedding dimension = {hidden_size} should be divisible "
            f"by number of heads = {num_heads}"
        )
    initializer = {
        "kernel_initializer": TruncatedNormal(stddev=0.02),
        "bias_initializer": "zeros",
    }
    query = Dense(hidden_size, name=f"{name}/query", **initializer)(x)
    key = Dense(hidden_size, name=f"{name}/key", **initializer)(x)
    value = Dense(hidden_size, name=f"{name}/value", **initializer)(x)
    attention, weights = Attention(num_heads=num_heads, softmax=softmax,
                                   name=f"{name}/attention")([query, key, value])
    output = Dense(hidden_size, name=f"{name}/out", **initializer)(attention)
    return output, weights


def transformer_block(inputs, num_heads, hidden_size, mlp_dim, dropout, name,
                      norm='LN', softmax='softmax', mlp_act="GeLU"):
    """Transformer block definition.

    Args:
        inputs (tf.Tensor): inputs
        num_heads (int): the number of attention heads
        hidden_size (int): multi-head attention block internal size
        mlp_dim (int): MLP block internal size
        dropout (float): dropout rate
        name (str): used as a base name for the layers in the block
        norm (str, optional): string that values in ['LN', 'GN1', 'BN', 'LMN'] and that allows to
            choose from LayerNormalization, GroupNormalization(groups=1, ...), BatchNormalization
            or LayerMadNormalization layers respectively in the block. Defaults to 'LN'.
        softmax (str, optional): string with values in ['softmax', 'softmax2'] that allows to choose
            between softmax and softmax2 activation in attention. Defaults to 'softmax'.
        mlp_act (str, optional): string that values in ['GeLU', 'ReLUx', 'swish'] and that allows to
            choose from GeLU, ReLUx or swish activation in the MLP block. Defaults to "GeLU".

    Returns:
        (tf.Tensor, (tf.Tensor, tf.Tensor)): block outputs and (attention softmaxed scores, the
        normalized sum of inputs and attention outputs)
    """
    x = norm_to_layer(norm)(epsilon=1e-6, name=f"{name}/LayerNorm_0")(inputs)
    x, weights = multi_head_attention(x,
                                      num_heads=num_heads,
                                      hidden_size=hidden_size,
                                      name=f"{name}/MultiHeadDotProductAttention_1",
                                      softmax=softmax
                                      )
    x = Dropout(dropout)(x)
    x_norm2 = Add(name=f"{name}/add_1")([x, inputs])
    y = norm_to_layer(norm)(epsilon=1e-6, name=f"{name}/LayerNorm_2")(x_norm2)
    y = mlp_block(y, mlp_dim, dropout, f"{name}/MlpBlock", mlp_act)
    outputs = Add(name=f"{name}/add_2")([x_norm2, y])
    return outputs, (weights, x_norm2)


def conv_transpose_block(inputs,
                         filters,
                         kernel_size,
                         add_batchnorm=False,
                         add_activation=True,
                         **kwargs):
    """Adds a transposed convolutional layer with optional layers in the following order:
    batch normalization, activation.

    Args:
        inputs (tf.Tensor): input tensor of shape `(rows, cols, channels)`
        filters (int): the dimensionality of the output space (i.e. the number of output filters in
            the convolution).
        kernel_size (int or tuple of 2 integers): specifying the height and width of the 2D
            convolution kernel. Can be a single integer to specify the same value for all spatial
            dimensions.
        add_batchnorm (bool, optional): add a BatchNormalization layer. Defaults to False.
        add_activation (bool, optional): add a ReLU8 layer. Defaults to True.
        **kwargs: arguments passed to the keras.Conv2DTranspose layer, such as strides, padding,
            use_bias, weight_regularizer, etc.

    Returns:
        tf.Tensor: output tensor of transposed convolution block.
    """
    if 'activation' in kwargs and kwargs['activation']:
        raise ValueError("Keyword argument 'activation' in conv_transpose_block must be None.")
    if 'dilation_rate' in kwargs and kwargs['dilation_rate'] not in [1, [1, 1], (1, 1)]:
        raise ValueError("Keyword argument 'dilation_rate' is not supported in "
                         "conv_transpose_block.")

    conv_trans_layer = Conv2DTranspose(filters, kernel_size, **kwargs)
    x = conv_trans_layer(inputs)

    if add_batchnorm:
        x = BatchNormalization(name=conv_trans_layer.name + '/BN')(x)

    if add_activation:
        x = ReLU(8.0, name=conv_trans_layer.name + '/relu')(x)

    return x


def sepconv_transpose_block(inputs,
                            filters,
                            kernel_size,
                            strides=2,
                            padding='same',
                            use_bias=True,
                            add_batchnorm=False,
                            add_activation=True,
                            name=None,
                            kernel_initializer='glorot_uniform',
                            pointwise_regularizer=None):
    """Adds a transposed separable convolutional layer with optional layers in the following order:
    batch normalization, activation.

    The separable operation is made of a DepthwiseConv2DTranspose followed by a pointwise Conv2D.

    Args:
        inputs (tf.Tensor): input tensor of shape `(rows, cols, channels)`
        filters (int): the dimensionality of the output space (i.e. the number of output filters in
            the pointwise convolution).
        kernel_size (int or tuple of 2 integers): specifying the height and width of the depthwise
            transpose kernel. Can be a single integer to specify the same value for all spatial
            dimensions.
        strides (int or tuple of 2 integers, optional): strides of the transposed depthwise.
            Defaults to 2.
        padding (str, optional): padding mode for the transposed depthwise. Defaults to 'same'.
        use_bias (bool, optional): whether the layer uses a bias vectors. Defaults to True.
        add_batchnorm (bool, optional): add a BatchNormalization layer. Defaults to False.
        add_activation (bool, optional): add a ReLU8 layer. Defaults to True.
        name (str, optional): name of the layer. Defaults to None.
        kernel_initializer (keras.initializer, optional): initializer for both kernels. Defaults to
            'glorot_uniform'.
        pointwise_regularizer (keras.regularizers, optional): regularizer function applied to the
            pointwise kernel matrix. Defaults to None.

    Returns:
        tf.Tensor: output tensor of transposed separable convolution block.
    """
    if name:
        dw_name = "dw_" + name
        pw_name = "pw_" + name
    else:
        dw_name, pw_name = None, None

    dw_trans_layer = DepthwiseConv2DTranspose(kernel_size,
                                              strides=strides,
                                              padding=padding,
                                              use_bias=use_bias,
                                              depthwise_initializer=kernel_initializer,
                                              name=dw_name)
    pw_layer = Conv2D(filters,
                      kernel_size=1,
                      padding='valid',
                      use_bias=use_bias,
                      kernel_regularizer=pointwise_regularizer,
                      kernel_initializer=kernel_initializer,
                      name=pw_name)

    x = dw_trans_layer(inputs)
    x = pw_layer(x)

    if add_batchnorm:
        x = BatchNormalization(name=pw_layer.name + '/BN')(x)

    if add_activation:
        x = ReLU(8.0, name=pw_layer.name + '/relu')(x)

    return x


def yolo_head_block(x, num_boxes, classes, filters=1024):
    """Adds the `YOLOv2 detection head <https://arxiv.org/pdf/1612.08242.pdf>`_, at the output
    of a model.

    Args:
        x (:obj:`tf.Tensor`): input tensor of shape `(rows, cols, channels)`.
        num_boxes (int): number of boxes.
        classes (int): number of classes.
        filters (int, optional): number of filters in hidden layers. Defaults to 1024.

    Returns:
        :obj:`tf.Tensor`: output tensor of yolo detection head block.

    Notes:
        This block replaces conv layers by separable_conv, to decrease the amount of parameters.
    """
    x = separable_conv_block(x, filters=filters, name='1conv',
                             kernel_size=(3, 3), padding='same', use_bias=False,
                             add_activation=True, add_batchnorm=True)
    x = separable_conv_block(x, filters=filters, name='2conv',
                             kernel_size=(3, 3), padding='same', use_bias=False,
                             add_activation=True, add_batchnorm=True)
    x = separable_conv_block(x, filters=filters, name='3conv',
                             kernel_size=(3, 3), padding='same', use_bias=False,
                             add_activation=True, add_batchnorm=True)
    x = separable_conv_block(x, filters=(num_boxes * (4 + 1 + classes)), name='detection_layer',
                             kernel_size=(3, 3), padding='same', use_bias=True,
                             add_activation=False, add_batchnorm=False)
    return x


def vit_extract_feature_map(y, feat_shape, num_non_patch=1):
    """Add a ExtractToken + Reshape layers to convert the plain ViT output into a feature map.

    All tokens will be extracted except those considered as class tokens (``num_non_patch``)

    Args:
        y (:obj:`tf.Tensor`): output of ViT classifier model.
        feat_shape (tuple of int): height/width desired output size
        num_non_patch (int, optional): number of non-patch tokens to exclude on the final output.
            Defaults to 1.

    Returns:
        :obj:`keras.Model`: ViT with new layers
    """
    y = ExtractToken(name="extract_features", token=list(range(num_non_patch, y.shape[1])))(y)
    y = Reshape(target_shape=(*feat_shape, y.shape[-1]), name="features")(y)
    return y



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
This module provides some utils to handle boxes and calculate the jaccard overlap score.
"""
import numpy as np
import tensorflow as tf


def xywh_to_xyxy(boxes):
    """Convert a set of boxes from format xywh to xyxy, where each format represent:

        * 'xywh': format of ('cx', 'xy', 'w', 'h'), also called 'centroids' and
        * 'xyxy': format of ('x_min', 'y_min', 'x_max', 'y_max'), also called 'corners'.

    Args:
        boxes (tf.tensor or np.ndarray): tensor with shape (N, 4)

    Returns:
        tf.tensor or np.ndarray: tensor with new format
    """
    assert boxes.shape[-1] == 4, "Expected 4 as last dimension."
    x1y1 = boxes[..., :2] - 0.5 * boxes[..., -2:]
    x2y2 = boxes[..., :2] + 0.5 * boxes[..., -2:]
    if isinstance(boxes, tf.Tensor):
        return tf.concat([x1y1, x2y2], axis=-1)
    return np.concatenate([x1y1, x2y2], axis=-1)


def xyxy_to_xywh(boxes):
    """Convert a set of boxes from format xyxy to xywh, where each format represent:

        * 'xyxy': format of ('x_min', 'y_min', 'x_max', 'y_max'), also called 'corners' and
        * 'xywh': format of ('cx', 'xy', 'w', 'h'), also called 'centroids'.

    Args:
        boxes (tf.tensor): tensor with shape (N, 4)

    Returns:
        tf.tensor: tensor with new format
    """
    assert boxes.shape[-1] == 4, "Expected 4 as last dimension."
    tf.debugging.assert_greater_equal(boxes[..., -2:], boxes[..., :2],
                                      "x2y2 must be greater than x1y1")
    wh = boxes[..., -2:] - boxes[..., :2]
    xy = boxes[..., :2] + 0.5 * wh
    if isinstance(boxes, tf.Tensor):
        return tf.concat([xy, wh], axis=-1)
    return np.concatenate([xy, wh], axis=-1)


def compute_overlap(a1, a2, mode="element_wise", box_format="xywh"):
    """Calculate ious between a1, a2 in two different modes:

        * element_wise: compute iou element-by-element, returning 1D array tensor,
        * outer_product: compute cross iou with all possible combination between inputs.

    Args:
        a1 (tf.tensor or np.ndarray): set of boxes, with shape at least equal to (N, 4).
        a2 (tf.tensor or np.ndarray): set of boxes, with compatible broadcast-shape
            (in 'element_wise' mode) or shape at least equal to (N, 4) (in 'outer_product' mode).
        box_format (str, optional): format of both inputs. Defaults to 'xywh'.

    Returns:
        tf.tensor or np.ndarray: IoU between inputs with shape (N,) in 'element_wise',
        otherwise (N, M).
    """
    format_choices = ["xywh", "xyxy"]
    mode_choises = ["element_wise", "outer_product"]
    assert box_format in format_choices, f"box format must be one of {format_choices}"
    assert mode in mode_choises, f"mode must be one of {mode_choises}"

    def process_boxes(a):
        # Decompose the input set of boxes into three parts:
        # [(x_min, y_min), (x_max, y_max), (w, h)]
        if box_format == "xywh":
            a_wh = a[..., 2:4]
            a = xywh_to_xyxy(a[..., :4])
        else:
            a_wh = a[..., 2:4] - a[..., :2]
        return a[..., :2], a[..., 2:4], a_wh

    # Check inputs are not empty
    a1_num_boxes = a1.shape[0]
    a2_num_boxes = a2.shape[0]
    is_tf_tensor = isinstance(a1, tf.Tensor)
    if a1_num_boxes == 0 or a2_num_boxes == 0:
        if is_tf_tensor:
            return tf.zeros((a1_num_boxes, a2_num_boxes))
        return np.zeros((a1_num_boxes, a2_num_boxes))

    # In outer product, we need expand tensors. Broadcast will do the rest of the job.
    if mode == "outer_product":
        a1 = a1[:, None]
        a2 = a2[None]

    # Process two sets
    a2_mins, a2_maxes, a2_wh = process_boxes(a2)
    a1_mins, a1_maxes, a1_wh = process_boxes(a1)

    # Intersection as min((a2_maxes, a1_maxes) - max(a2_mins, a1_mins)
    intersect_mins = tf.math.maximum(a2_mins, a1_mins)
    intersect_maxes = tf.math.minimum(a2_maxes, a1_maxes)

    # Getting the intersections in the xy (aka the width, height intersection)
    intersect_wh = tf.math.maximum(intersect_maxes - intersect_mins, 0)

    # Multiply to get intersecting area
    intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

    # Values for the single sets
    true_areas = a1_wh[..., 0] * a1_wh[..., 1]
    pred_areas = a2_wh[..., 0] * a2_wh[..., 1]

    # Compute union for the IoU
    add_areas = tf.convert_to_tensor(pred_areas + true_areas, dtype=intersect_areas.dtype)
    union_areas = add_areas - intersect_areas
    iou = intersect_areas / union_areas

    if is_tf_tensor:
        return iou
    return iou.numpy()



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2023 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Processing tools for CenterNet data handling.
"""
import numpy as np
from tensorflow.nn import max_pool2d

from ..detection.processing import BoundingBox


def decode_output(output,
                  nb_classes,
                  obj_threshold=0.1,
                  max_detections=100,
                  kernel=5):
    """ Decodes a CenterNet model.

    Args:
        output (tf.Tensor): model output to decode.
        nb_classes (int): number of classes.
        obj_threshold (float, optional): confidence threshold for a box. Defaults to 0.1.
        max_detection (int, optional): maximum number of boxes the model is allowed to produce.
            Defaults to 100.
        kernel (int, optional) : max pool kernel size. Defaults to 5.

    Returns:
        List: `BoundingBox` objects
    """

    def _sigmoid(x):
        return 1. / (1. + np.exp(-x))

    grid_h, grid_w = output.shape[:2]

    # Decode the output of the network
    center_heatmap_pred = _sigmoid(output[..., :nb_classes])
    wh_pred = output[..., nb_classes:nb_classes + 2]
    offset_pred = output[..., nb_classes + 2:nb_classes + 4]

    # Get local maximum
    hmax = max_pool2d(center_heatmap_pred[None, ...],
                      ksize=[kernel, kernel], strides=1, padding='SAME', data_format='NHWC')
    center_heatmap_pred[hmax[0] != center_heatmap_pred] = 0

    # Get top k from the heatmap
    perm_center_heatmap = np.transpose(center_heatmap_pred, (2, 0, 1))
    flattened_heatmap = np.reshape(perm_center_heatmap, (-1))

    topk_scores = np.partition(flattened_heatmap, -max_detections)[-max_detections:]
    topk_scores = np.flip(np.sort(topk_scores))
    topk_inds = np.argpartition(flattened_heatmap, -max_detections)[-max_detections:]
    topk_inds = topk_inds[np.argsort(flattened_heatmap[topk_inds])][::-1]
    topk_labels = topk_inds // (grid_h * grid_w)
    topk_inds = topk_inds % (grid_h * grid_w)
    topk_ys = topk_inds // grid_h
    topk_xs = topk_inds % grid_w

    # Transpose and gather features for the WH and OFFSET.
    # Removed the transpose as we don't do it above either
    wh_pred = np.reshape(wh_pred, [-1, wh_pred.shape[-1]])
    wh = wh_pred[topk_inds, ...]
    offset_pred = np.reshape(offset_pred, [-1, offset_pred.shape[-1]])
    offset = offset_pred[topk_inds, ...]

    # The output should be x,y,w,h
    topk_xs = topk_xs + offset[..., 0]
    topk_ys = topk_ys + offset[..., 1]
    tl_x = np.clip((topk_xs - wh[..., 0] / 2) / grid_w, a_min=0, a_max=grid_w)
    tl_y = np.clip((topk_ys - wh[..., 1] / 2) / grid_h, a_min=0, a_max=grid_h)
    br_x = np.clip((topk_xs + wh[..., 0] / 2) / grid_w, a_min=0, a_max=grid_w)
    br_y = np.clip((topk_ys + wh[..., 1] / 2) / grid_h, a_min=0, a_max=grid_h)

    boxes = []
    for i in range(max_detections):
        score = topk_scores[i]
        if score > obj_threshold:
            label = topk_labels[i]
            box = BoundingBox(tl_x[i], tl_y[i], br_x[i], br_y[i], score=score)
            box.label = label
            boxes.append(box)
        else:
            break

    return boxes



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2021 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
Common utility methods used by the training scripts.
"""

import os.path
import tempfile
import argparse
import numpy as np

from keras.preprocessing.image import DirectoryIterator
from keras.optimizers import Adam
from keras.layers import Activation
from keras.callbacks import ModelCheckpoint
from keras.utils.generic_utils import Progbar
from quantizeml.layers import calibration


def compile_model(model,
                  learning_rate=1e-3,
                  loss='categorical_crossentropy',
                  metrics=None):
    """ Compiles the model using Adam optimizer.

    Args:
        model (keras.Model): the model to compile
        learning_rate (float, optional): the learning rate. Defaults to 1e-3.
        loss (str or function, optional): the loss function. Defaults to
            'categorical_crossentropy'.
        metrics (list, optional): list of metrics to be evaluated during
            training and testing. Defaults to None.
    """
    if metrics is None:
        metrics = ['accuracy']

    model.compile(optimizer=Adam(learning_rate=learning_rate),
                  loss=loss,
                  metrics=metrics)


def freeze_model_before(model, freeze_before):
    """ Freezes the model before the given layer name.

    Args:
        model (keras.Model): the model to freeze
        freeze_before (str): name of the layer from which the model will not be
            frozen

    Raises:
        ValueError: if the provided layer name was not found in the model
    """
    if freeze_before is not None:
        trainable = False
        for layer in model.layers:
            if layer.name == freeze_before:
                trainable = True
            layer.trainable = trainable
        if not trainable:
            raise ValueError(f"No such layer {freeze_before} in model.")


def calibrate_model(model, x, num_samples, batch_size=None, epochs=1):
    """Calibrates model

    Args:
        model (keras.Model): the model to calibrate
        x (tf.Dataset, np.array or generator): train input data
        num_samples (int): total number of samples before declaring the calibration round finished.
        batch_size (int, optional): the batch size. Defaults to None.
        epochs (int, optional): the number of epochs. Defaults to 1.
    """
    if batch_size is None:
        steps = num_samples
    else:
        assert batch_size > 0, "The batch size should be strictly positive."
        steps = np.ceil(num_samples / batch_size)

    with calibration(True):
        progbar = Progbar(epochs)
        for epoch in range(epochs):
            progbar.update(epoch)
            model.predict(x=x, steps=steps, batch_size=batch_size, verbose=0)
        progbar.update(epochs, finalize=epochs)


def evaluate_model(model,
                   x,
                   y=None,
                   batch_size=None,
                   steps=None,
                   print_history=False):
    """ Evaluates model performances.

    Args:
        model (keras.Model): the model to evaluate
        x (tf.Dataset, np.array or generator): evaluation input data
        y (tf.Dataset, np.array or generator, optional): evaluation target
            data. Defaults to None.
        batch_size (int, optional): the batch size. Defaults to None.
        steps (int, optional): total number of steps before declaring the
            evaluation round finished. Defaults to None.
        print_history (bool, optional): either to print all history or just
            accuracy. Defaults to False.
    """
    history = model.evaluate(x=x,
                             y=y,
                             batch_size=batch_size,
                             steps=steps,
                             verbose=0)

    if print_history:
        print('Validation history: ', history)
    else:
        print('Validation accuracy: ', history[1])


def evaluate_akida_model(model, x, activation):
    """ Evaluates Akida model and return predictions and labels to compute
    accuracy.

    Args:
        model (akida.Model): the model to evaluate
        x (tf.Dataset, np.array or generator): evaluation input data
        activation (str): activation function to apply to potentials

    Returns:
        np.array, np.array: predictions and labels
    """

    # Initialize to None to allow different shapes depending on the caller
    labels = None
    pots = None

    # Datasets created as ImageDataGenerator will be looped over indefinitely,
    # they thus need a special treatment
    steps = None
    if isinstance(x, DirectoryIterator):
        steps = np.math.ceil(x.samples / x.batch_size)

    for batch, label_batch in x:
        if not isinstance(batch, np.ndarray):
            batch = batch.numpy()

        pots_batch = model.predict(batch.astype('uint8'))

        if labels is None:
            labels = label_batch
            pots = pots_batch.squeeze(axis=(1, 2))
        else:
            labels = np.concatenate((labels, label_batch))
            pots = np.concatenate((pots, pots_batch.squeeze(axis=(1, 2))))

        # End the for loop if the number of batches has reached the number of
        # steps
        if steps is not None and x.total_batches_seen == steps:
            break

    return Activation(activation)(pots), labels


def get_training_parser(batch_size,
                        calibrate=False,
                        tune=False,
                        freeze_before=False,
                        global_batch_size=True,
                        global_parser=None):
    """ Instantiates a base arguments parser for training scripts.

    The returned parser comes with train, tune and eval actions as subparsers
    (with default parameters) for which one can add arguments if required.

    Args:
        batch_size (int): batch size default value
        calibrate (bool, optional): either to add a calibrate action or not. Defaults to
            False.
        tune (bool, optional): either to add a tune action or not. Defaults to
            False.
        freeze_before (bool, optional): either to add a "--freeze_before"
            parameter in train and tune actions or not. Defaults to False.
        global_batch_size (bool, optional): either to add batch size as a
            global parser parameter or for train/tune actions only. Defaults to
            True.
        global_parser (argparse.ArgumentParser, optional): global parser with
            custom parameters. Defaults to None.

    Returns:
        argparse.ArgumentParser, argparse.ArgumentParser, argparse.ArgumentParser,
        argparse.ArgumentParser, argparse.ArgumentParser, argparse._SubParsersAction: main parser
        and train, calibrate, tune and eval actions subparsers and the subparser object.
    """

    # Define a strictly positive int type for parameters
    def positive_int(value):
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError(
                "%s is an invalid value, must be >0." % value)
        return ivalue

    parser = argparse.ArgumentParser()

    # Base arguments
    if global_parser is None:
        global_parser = argparse.ArgumentParser(add_help=False)

    global_parser.add_argument("-m",
                               "--model",
                               type=str,
                               required=True,
                               help="Model to load")
    if global_batch_size:
        global_parser.add_argument("-b",
                                   "--batch_size",
                                   type=positive_int,
                                   default=batch_size,
                                   help="The batch size")

    # Create an action subparser
    subparsers = parser.add_subparsers(help="Desired action to perform",
                                       dest="action")

    # Create parent subparser for arguments shared between train and tune
    # actions
    parent_parser = argparse.ArgumentParser(add_help=False,
                                            parents=[global_parser])
    parent_parser.add_argument("-s",
                               "--savemodel",
                               type=str,
                               default=None,
                               help="Save model with the specified name")
    parent_parser.add_argument("-e",
                               "--epochs",
                               type=positive_int,
                               required=True,
                               help="The number of training epochs")
    if not global_batch_size:
        parent_parser.add_argument("-b",
                                   "--batch_size",
                                   type=positive_int,
                                   default=batch_size,
                                   help="The batch size")
    if freeze_before:
        parent_parser.add_argument(
            "-fb",
            "--freeze_before",
            type=str,
            default=None,
            help="Freeze the layers of the model before this \
                                layer")

    # Subparsers based on parent
    train_parser = subparsers.add_parser("train",
                                         parents=[parent_parser],
                                         help="Train a Keras model")
    tune_parser = None
    if tune:
        tune_parser = subparsers.add_parser("tune",
                                            parents=[parent_parser],
                                            help="Tune a Keras model")

    # Other subparsers
    calibrate_parser = None
    if calibrate:
        calibrate_parser = subparsers.add_parser("calibrate",
                                                 parents=[parent_parser],
                                                 help="Calibrate a Keras model")
        calibrate_parser.add_argument("-ns",
                                      "--num_samples",
                                      type=positive_int,
                                      default=1024,
                                      help="Number of samples to use for calibration")

    eval_parser = subparsers.add_parser("eval",
                                        help="Evaluate a model",
                                        parents=[global_parser])

    eval_parser.add_argument("-ak",
                             "--akida",
                             action='store_true',
                             help="Converts to an Akida model and evaluate")

    return parser, train_parser, tune_parser, eval_parser, calibrate_parser, subparsers


def print_history_stats(history):
    """Prints a basic statistics of the training/eval history.

    Args:
        history (keras.callbacks.History): history of the training/eval process
    """
    print("\nHistory statistics:\n" + "-" * 80 + "\n")
    max_len_metric = max(len(metric) for metric in history.history.keys())
    clean_history = {key.ljust(max_len_metric): values for key, values in history.history.items()}
    stats = [np.min, np.max, np.mean, np.std]
    print("Parameter:".ljust(max_len_metric) + "\t" + "\t".join(f.__name__.ljust(6) for f in stats))
    for key, values in clean_history.items():
        values = "\t".join(f"{f(values):.4f}" for f in stats)
        print(f"{key}\t{values}")


class RestoreBest(ModelCheckpoint):
    """Keep best weights on training

    This tool uses a ModelCheckPoint callback to save the best weights during the training and
    restore them in the model when training is complete. The best model is saved to a
    temporary directory. So if the training does not go to the end, this model can be used for
    start a new one.

    Args:
        model (keras.Model): the model to load the best weights
        monitor (str, optional): the metric name to monitor. Defaults to 'val_accuracy'.
        mode (str, optional): one of {'auto', 'min', 'max'}. The decision to overwrite the current
            save model is made based on either the maximization or the minimization of the monitored
            quantity. Defaults to 'max'.
    """
    def __init__(self, model, monitor='val_accuracy', mode='max'):
        self.model = model
        self.temp_dir = tempfile.TemporaryDirectory()
        self.filepath = os.path.join(self.temp_dir.name, "best_model.h5")
        print(f"Best model saved as {self.filepath}.")
        super().__init__(filepath=self.filepath,
                         monitor=monitor,
                         save_best_only=True,
                         mode=mode)

    def on_train_end(self, logs=None):
        if os.path.exists(self.filepath):
            self.model.load_weights(self.filepath)
        # Clean temp directory
        self.temp_dir.cleanup()


def save_model(model, model_path, save_path=None, action_str='saved'):
    """ Saves the model with the given or a default name.

    Args:
        model (keras.Model): the model to save
        model_path (str): the original model path
        save_path (str): the path to save the model. Defaults to None.
        action_str (str): the 'action' used in the training script, will be appended to the original
            model path when save_path is None. Defaults to 'saved'.
    """
    if not save_path:
        save_path = os.path.splitext(model_path)[0]
        save_path += f"_{action_str}.h5"

    model.save(save_path, include_optimizer=False)
    print(f"Model saved as {save_path}.")



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2020 Brainchip Holdings Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
"""
akida_models main command-line interface

This package entry-point allows akida models to be instantiated and saved at a
specified location.

"""

import argparse
import os

from .utk_face.model_vgg import vgg_utk_face
from .kws.model_ds_cnn import ds_cnn_kws
from .modelnet40.model_pointnet_plus import pointnet_plus_modelnet40
from .imagenet.model_mobilenet import mobilenet_imagenet
from .imagenet.model_akidanet import akidanet_imagenet
from .imagenet.model_akidanet_edge import akidanet_edge_imagenet
from .imagenet.model_akidanet18 import akidanet18_imagenet
from .detection.model_yolo import yolo_base
from .dvs.model_convtiny_handy import convtiny_dvs_handy
from .dvs.model_convtiny_gesture import convtiny_dvs_gesture
from .mnist.model_gxnor import gxnor_mnist
from .transformers.model_vit import vit_ti16, bc_vit_ti16
from .transformers.model_deit import deit_ti16, bc_deit_ti16
from .portrait128.model_akida_unet import akida_unet_portrait128
from .centernet.model_centernet import centernet_base
from .macs import display_macs
from .utils import load_weights, save_weights, load_model
from .unfuse_sepconv_layers import unfuse_sepconv2d


def save_model_weights(model_path, weights_path):
    """ CLI entry point to save the model weights in an npz file.

    Args:
        model_path (str): Path to the model to extract the weights from.
        weights_path (str): Path to save the npz file.
            Defaults to <model_path>.npz.
    """
    # Build name for weights file
    if weights_path is None:
        model_name = os.path.splitext(model_path)[0]
        weights_path = f"{model_name}.npz"

    # Load the model and save its weights
    model = load_model(model_path)
    save_weights(model, weights_path)
    print(f"Saved model weights to {weights_path}.")


def load_model_weights(model_path, weights_path):
    """ CLI entry point to apply weights to a model from an npz file.

    Args:
        model_path (str): Path to the model on which apply the weights.
        weights_path (str): Path to load the npz file.
    """
    # Update the model weights with the npz file
    model = load_model(model_path)
    load_weights(model, weights_path)
    model.save(model_path, include_optimizer=False)
    print(f"Saved model with new weights to {model_path}.")


def add_transformers_args(parser, add_customs=False, default_norm='LN', default_act='GeLU'):
    """ Add transformers commons arguments to the given parser.

    Args:
        parser (argparse.ArgumentParser): parser to add args to
        add_customs (bool, optional): indicates if normalization, softmax,
            last normalization and activation can be customized. Defaults to
            False.
        default_norm (str, optional): when add_customs=True, default value for normalization.
            Defaults to 'LN'.
        default_act (str, optional): when add_customs=True, default value for activation. Defaults
            to 'GeLU'.
    """
    parser.add_argument("-c", "--classes", type=int, default=1000,
                        help="The number of classes, by default %(default)s.")
    parser.add_argument("-bw", "--base_weights", type=str, default=None,
                        help="Optional keras weights to load in the model, by default %(default)s.")
    if add_customs:
        norms = ['LN', 'GN1', 'BN', 'LMN']
        parser.add_argument("--norm", type=str, choices=norms, default=default_norm,
                            help="Replace normalization in model with a custom function, by \
                                  default %(default)s")
        last_norms = ['LN', 'BN']
        parser.add_argument("--last_norm", type=str, choices=last_norms,
                            default=default_norm,
                            help=f"Replace last normalization in model with a \
                                  custom function, by default {default_norm}")
        sms = ['softmax', 'softmax2']
        parser.add_argument("--softmax", type=str, choices=sms, default='softmax',
                            help="Replace softmax operation in model with custom function, by \
                                  default %(default)s")
        acts = ['GeLU', 'ReLU8', 'swish']
        parser.add_argument("--act", type=str, choices=acts, default=default_act,
                            help="Replace activation function in model with custom function, \
                                  by default %(default)s")


def add_common_args(parser, default_classes, default_alpha=None, default_img_size=None,
                    bw_required=None):
    """ Add commons arguments to the given parser.

    Args:
        parser (argparse.ArgumentParser): parser to add args to.
        default_classes (int): default number of classes.
        default_alpha (float, optional): default alpha. Defaults to None.
        default_img_size (int, optional): default image size. Defaults to None.
        bw_required (bool, optional): when base_weights is required. Defaults to None.
    """
    parser.add_argument("-c", "--classes", type=int, default=default_classes,
                        help="The number of classes, by default %(default)s.")
    if default_img_size is not None:
        img_sizes = [32, 64, 96, 128, 160, 192, 224]
        parser.add_argument("-i", "--image_size", type=int,
                            default=default_img_size, choices=img_sizes,
                            help="The square input image size, by default %(default)s.")
    if default_alpha is not None:
        parser.add_argument("-a", "--alpha", type=float, default=default_alpha,
                            help="The width of the model, by default %(default)s.")

    if bw_required is not None:
        parser.add_argument("-bw", "--base_weights", type=str, required=bw_required,
                            help="The base weights to load on model.")


def main():
    """ CLI entry point.

    Contains an argument parser with specific arguments depending on the model
    to be created. Complete arguments lists available using the -h or --help
    argument.

    """
    parser = argparse.ArgumentParser()
    sp = parser.add_subparsers(dest="action")
    c_parser = sp.add_parser("create", help="Create an akida Keras model")
    c_parser.add_argument("-s",
                          "--save_model",
                          type=str,
                          default=None,
                          help="The path/name to use to save the model")
    csp = c_parser.add_subparsers(dest="model",
                                  help="The type of model to be instantiated")
    csp.add_parser("vgg_utk_face", help="A VGG-like UTKFace model")
    csp.add_parser("convtiny_dvs_handy", help="A Convtiny DVS handy model")

    dvsh_parser = csp.add_parser("convtiny_dvs_handy",
                                 help="A Convtiny DVS model for akida")
    add_common_args(dvsh_parser, default_classes=9)

    csp.add_parser("convtiny_dvs_gesture", help="A Convtiny DVS gesture model")

    dvsg_parser = csp.add_parser("convtiny_dvs_gesture",
                                 help="A Convtiny DVS model for akida")
    add_common_args(dvsg_parser, default_classes=10)

    csp.add_parser(
        "ds_cnn_kws",
        help="A Depthwise Separable MobileNet-like model for the Keyword"
        " Spotting example")

    mb_parser = csp.add_parser("mobilenet_imagenet",
                               help="A MobileNet V1 model for akida")
    add_common_args(mb_parser, default_classes=1000, default_alpha=1.0, default_img_size=224)

    an_parser = csp.add_parser("akidanet_imagenet", help="An AkidaNet model")
    add_common_args(an_parser, default_classes=1000, default_img_size=224, default_alpha=1.0)

    ane_parser = csp.add_parser("akidanet_edge_imagenet",
                                help="An AkidaNet model modified for akida \
                                edge learning")
    add_common_args(ane_parser, default_classes=1000, bw_required=True)

    an18_parser = csp.add_parser("akidanet18_imagenet", help="An AkidaNet18 model")
    add_common_args(an18_parser, default_classes=1000, default_img_size=224)

    pnet_parser = csp.add_parser("pointnet_plus_modelnet40",
                                 help="A PointNet++ model for Akida")
    add_common_args(pnet_parser, default_classes=40)
    pnet_parser.add_argument("-alpha",
                             "--alpha",
                             type=float,
                             default=1.0,
                             help="Network filters multiplier (typically 0.5 , \
                             [1.0]")

    yl_parser = csp.add_parser("yolo_base", help="A YOLOv2 model for detection")
    add_common_args(yl_parser, default_classes=1, default_alpha=0.5, bw_required=False)
    yl_parser.add_argument("-na",
                           "--number_anchors",
                           type=int,
                           default=5,
                           help="The number of anchor boxes")

    csp.add_parser("gxnor_mnist", help="A GXNOR MNIST model for akida")

    akun_parser = csp.add_parser("akida_unet_portrait128",
                                 help="An Akida U-Net model for Portrait128 segmentation.")
    akun_parser.add_argument("-bw", "--base_weights", type=str, default=None,
                             help="The base AkidaNet weights to use for the encoder.")

    cn_parser = csp.add_parser("centernet", help="A CenterNet model.")
    add_common_args(cn_parser, default_classes=2, default_alpha=0.5, default_img_size=224,
                    bw_required=False)

    vit_parser = csp.add_parser("vit_ti16", help="A ViT model")
    add_transformers_args(vit_parser, add_customs=True)
    vit_parser.add_argument("-i", "--image_size", type=int, default=224,
                            choices=[224, 384], help="The square input image size")

    bc_vit_parser = csp.add_parser("bc_vit_ti16", help="A ViT, brainchip customized model")
    add_transformers_args(bc_vit_parser)

    deit_parser = csp.add_parser("deit_ti16", help="A DeiT model")
    add_transformers_args(deit_parser, add_customs=True)
    deit_parser.add_argument("-i", "--image_size", type=int, default=224,
                             choices=[160, 224], help="The square input image size")
    deit_parser.add_argument("-d", "--dist", action="store_true",
                             help="Use the model with a distilled token")

    bc_deit_parser = csp.add_parser("bc_deit_ti16", help="A DeiT, brainchip customized model")
    add_transformers_args(bc_deit_parser)
    bc_deit_parser.add_argument("-d", "--dist", action="store_true",
                                help="Use the model with a distilled token")
    bc_deit_parser.add_argument("-b", "--blocks", type=int, default=12,
                                help="Number of transformer blocks")

    macs_parser = sp.add_parser("macs", help="Get MACS for a Keras model")
    macs_parser.add_argument("-m", "--model", type=str)
    macs_parser.add_argument("-v", "--verbose", action="store_true")

    unfuse_parser = sp.add_parser(
        "unfuse", help="Unfuse SeparableConv2D layers of a Keras model")
    unfuse_parser.add_argument("-m", "--model", type=str, required=True, help="Model to unfuse.")

    # Save weights arguments
    w_parser = sp.add_parser(
        "save_weights", help="Store model's weights in an npz.")
    w_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to extract weights from.")
    w_parser.add_argument("-w", "--weights_path", type=str,
                          help="Npz file that contains the saved weights.")

    # Load weights arguments
    w_parser = sp.add_parser(
        "load_weights", help="Apply weights to a model from an npz file.")
    w_parser.add_argument("-m", "--model", type=str, required=True,
                          help="Model to apply the weights.")
    w_parser.add_argument("-w", "--weights_path", type=str, required=True,
                          help="Npz file that contains the weights to apply.")

    args = parser.parse_args()
    if args.action == "create":
        # Instantiate the wished model
        if args.model == "vgg_utk_face":
            model = vgg_utk_face()
        elif args.model == "convtiny_dvs_handy":
            model = convtiny_dvs_handy(classes=args.classes)
        elif args.model == "convtiny_dvs_gesture":
            model = convtiny_dvs_gesture(classes=args.classes)
        elif args.model == "ds_cnn_kws":
            model = ds_cnn_kws()
        elif args.model == "pointnet_plus_modelnet40":
            model = pointnet_plus_modelnet40(alpha=args.alpha,
                                             classes=args.classes)
        elif args.model == "mobilenet_imagenet":
            input_shape = (args.image_size, args.image_size, 3)
            model = mobilenet_imagenet(input_shape,
                                       alpha=args.alpha,
                                       classes=args.classes)
        elif args.model == "akidanet_imagenet":
            input_shape = (args.image_size, args.image_size, 3)
            model = akidanet_imagenet(input_shape,
                                      alpha=args.alpha,
                                      classes=args.classes)
        elif args.model == "akidanet_edge_imagenet":
            model = akidanet_edge_imagenet(base_model=args.base_weights,
                                           classes=args.classes)
        elif args.model == "akidanet18_imagenet":
            input_shape = (args.image_size, args.image_size, 3)
            model = akidanet18_imagenet(input_shape, classes=args.classes)
        elif args.model == "yolo_base":
            model = yolo_base(classes=args.classes,
                              nb_box=args.number_anchors,
                              alpha=args.alpha)
            if args.base_weights is not None:
                model.load_weights(args.base_weights, by_name=True)
        elif args.model == "centernet":
            model = centernet_base(input_shape=(args.image_size, args.image_size, 3),
                                   classes=args.classes, alpha=args.alpha)
            if args.base_weights is not None:
                model.load_weights(args.base_weights, by_name=True)
        elif args.model == "gxnor_mnist":
            model = gxnor_mnist()
        elif args.model == "akida_unet_portrait128":
            model = akida_unet_portrait128()
            if args.base_weights is not None:
                model.load_weights(args.base_weights, by_name=True)
        else:
            # Handle transformer models
            if args.model == "vit_ti16":
                input_shape = (args.image_size, args.image_size, 3)
                model = vit_ti16(input_shape=input_shape,
                                 classes=args.classes,
                                 norm=args.norm,
                                 last_norm=args.last_norm,
                                 softmax=args.softmax,
                                 act=args.act)
            elif args.model == "bc_vit_ti16":
                model = bc_vit_ti16(classes=args.classes)
            elif args.model == "deit_ti16":
                input_shape = (args.image_size, args.image_size, 3)
                model = deit_ti16(input_shape=input_shape,
                                  classes=args.classes,
                                  distilled=args.dist,
                                  norm=args.norm,
                                  last_norm=args.last_norm,
                                  softmax=args.softmax,
                                  act=args.act)
            elif args.model == "bc_deit_ti16":
                model = bc_deit_ti16(classes=args.classes, distilled=args.dist,
                                     num_blocks=args.blocks)
            # Load provided weights
            if args.base_weights is not None:
                model.load_weights(args.base_weights, by_name=True)
        # No need for default behaviour as the command-line parser only accepts
        # valid model types
        model_path = args.save_model
        if model_path is None:
            models_by_name = ("mobilenet_imagenet", "centernet")
            model_path = f"{model.name}.h5" if args.model in models_by_name else f"{args.model}.h5"

        # If needed, add the extension
        if not model_path.endswith(".h5"):
            model_path = f"{model_path}.h5"
        model.save(model_path, include_optimizer=False)
        print(f"Model saved as {model_path}")

    elif args.action == "macs":
        display_macs(args.model, args.verbose)
    elif args.action == "save_weights":
        save_model_weights(
            model_path=args.model,
            weights_path=args.weights_path,
        )
    elif args.action == "load_weights":
        load_model_weights(
            model_path=args.model,
            weights_path=args.weights_path,
        )
    elif args.action == "unfuse":
        model = load_model(args.model)
        unfused_model = unfuse_sepconv2d(model)
        # strip file extension from model path
        model_path = os.path.splitext(args.model)[0]
        # Write new path for unfused model
        model_path += "_unfused.h5"
        # Save unfused model
        unfused_model.save(model_path, include_optimizer=False)
        print(f"{model.name} has been unfused and saved under {model_path}.")



import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import (
        convert_variables_to_constants_v2_as_graph)

from quantizeml.models import load_model
from cnn2snn import load_quantized_model


def get_flops(model):
    """
    Calculate FLOPS for a tf.keras.Model or tf.keras.Sequential model in inference mode.

    It uses tf.compat.v1.profiler under the hood.

    Args:
        model (:obj:`keras.Model`): the model to evaluate

    Returns:
        :obj:`tf.compat.v1.profiler.GraphNodeProto`: object containing the FLOPS
    """
    # Prepare a constant input to pass to the profiler
    input_shape = model.inputs[0].shape.as_list()
    input_shape[0] = 1
    x = tf.constant(tf.fill(input_shape, 1))
    if not isinstance(model, (tf.keras.models.Sequential, tf.keras.models.Model)):
        raise ValueError("Calculating FLOPS is only supported for `tf.keras.Model`"
                         "and `tf.keras.Sequential` instances.")

    # convert tf.keras model into frozen graph to count FLOPs about operations used at inference
    real_model = tf.function(model).get_concrete_function(x)
    frozen_func, _ = convert_variables_to_constants_v2_as_graph(real_model)
    # Calculate FLOPs with tf.profiler
    run_meta = tf.compat.v1.RunMetadata()
    opts = (
        tf.compat.v1.profiler.ProfileOptionBuilder(
            tf.compat.v1.profiler.ProfileOptionBuilder().float_operation()
        )
        .with_empty_output()
        .build()
    )

    flops = tf.compat.v1.profiler.profile(
        graph=frozen_func.graph, run_meta=run_meta, cmd="scope", options=opts
    )

    tf.compat.v1.reset_default_graph()

    return flops


def display_macs(model_path, verbose=False):
    """Displays the MACs for a keras model

    By default it displays only the total MACS.

    Args:
        model (:obj:`keras.Model`): the model to evaluate
        verbose (bool): display MACS for each operation

    """
    try:
        model = load_model(model_path)
    except ValueError:
        model = load_quantized_model(model_path)

    flops = get_flops(model)

    if verbose:
        def display_children_macs(nodes):
            for node in nodes:
                print(f"{node.name}: {node.total_float_ops / 2:e} MACS")
                display_children_macs(node.children)
        # Recursively display MACS by node (i.e. operation)
        display_children_macs(flops.children)

    # We divide FLOPS by 2 to obtain an estimate of Multiply and Accumulate (MACS)
    print(f"Total: {flops.total_float_ops / 2:e} MACS")



"""
Imports models.
"""

from .dvs.model_convtiny_gesture import (convtiny_dvs_gesture,
                                         convtiny_gesture_pretrained)
from .dvs.model_convtiny_handy import (convtiny_dvs_handy,
                                       convtiny_handy_samsung_pretrained)
from .imagenet.model_mobilenet import (mobilenet_imagenet,
                                       mobilenet_imagenet_pretrained)
from .imagenet.model_akidanet import (
    akidanet_imagenet, akidanet_imagenet_pretrained, akidanet_faceidentification_pretrained,
    akidanet_plantvillage_pretrained, akidanet_vww_pretrained)
from .imagenet.model_akidanet_edge import (
    akidanet_edge_imagenet, akidanet_edge_imagenet_pretrained,
    akidanet_faceidentification_edge_pretrained)
from .imagenet.model_akidanet18 import akidanet18_imagenet, akidanet18_imagenet_pretrained
from .kws.model_ds_cnn import ds_cnn_kws, ds_cnn_kws_pretrained
from .modelnet40.model_pointnet_plus import (pointnet_plus_modelnet40,
                                             pointnet_plus_modelnet40_pretrained
                                             )
from .utk_face.model_vgg import vgg_utk_face, vgg_utk_face_pretrained
from .detection.model_yolo import (yolo_base, yolo_widerface_pretrained,
                                   yolo_voc_pretrained)
from .centernet.model_centernet import centernet_base
from .mnist.model_gxnor import gxnor_mnist, gxnor_mnist_pretrained
from .transformers.model_vit import (
    vit_imagenet, vit_ti16, bc_vit_ti16, bc_vit_ti16_imagenet_pretrained, vit_s16, vit_s32, vit_b16,
    vit_b32, vit_l16, vit_l32)
from .transformers.model_deit import (
    deit_imagenet, deit_ti16, bc_deit_ti16, bc_deit_dist_ti16_imagenet_pretrained,
    deit_s16, deit_b16)
from .portrait128.model_akida_unet import akida_unet_portrait128

from .calibrable import calibration_required
from .gamma_constraint import add_gamma_constraint
from .filter_pruning import delete_filters, prune_model
from .utils import fetch_file
from .unfuse_sepconv_layers import unfuse_sepconv2d
